//---------------------------------------------------------------------------

#pragma hdrstop

#include "DirectSoundScope.h"
#include "ThreadComm.h"
#include "PluginSettings.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <algorithm>
#include <assert.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "dsound.lib"

DirectSoundScope scope;

/* DirectSound enum device callback */
static BOOL CALLBACK DSEnumCallback( LPGUID lpGuid, LPCSTR lpcstrDescription,  
				     LPCSTR lpcstrModule, LPVOID lpContext)
{
	unsigned index;
	
    /* Put the capture and playback of the same devices to the same 
     * dev_info item, by looking at the GUID.
     */
    for (index=0; index<scope.dev_count; ++index) {
	if ((scope.dev_info[index].lpGuid==NULL && lpGuid==NULL) ||
		memcmp(&scope.dev_info[index].guid, lpGuid, sizeof(GUID))==0)
	{
	    break;
	}
    }

	if (index == scope.dev_count)
		++scope.dev_count;
	else if (scope.dev_count >= DirectSoundScope::MAX_HARDWARE) {
		scope.log("Some of the devices are not listed");
		return FALSE;
    }

	scope.dev_info[index].name = lpcstrDescription;
	if (lpGuid == NULL) {
		scope.dev_info[index].lpGuid = NULL;
	} else {
		memcpy(&scope.dev_info[index].guid, lpGuid, sizeof(GUID));
		scope.dev_info[index].lpGuid = &scope.dev_info[index].guid;
    }

    return TRUE;
}

DirectSoundScope::DirectSoundScope():
	threadComm(NULL),
	hdr_bytes(0),
	hdr(NULL),
	bRunning(false),
	mode(TRIGGER_MODE_AUTO),
	lastTriggerDistance(0),
	averageA(0),
	averageB(0)
{
    bDeviceOpened = false;
	bStopRequest = false;
	callbackTrigger = NULL;
    callbackData = NULL;
    callbackLog = NULL;
	callbackConnect = NULL;

	settings.pChannelEnable = pChannelEnable;
	settings.eCouplingType = eCouplingType;
	settings.pSensitivity = pSensitivity;
	upcoming_settings.pChannelEnable = pChannelEnableUpcoming;
	upcoming_settings.eCouplingType = eCouplingTypeUpcoming;
	upcoming_settings.pSensitivity = pSensitivityUpcoming;

	// some default values
	settings.iChannelCount = CHANNEL_CNT;
	settings.pChannelEnable[0] = true;
	settings.pChannelEnable[1] = true;
	settings.iSamplingPeriod = 1;
	settings.eTriggerMode = TRIGGER_MODE_AUTO;
	settings.eTriggerType = TRIGGER_SLOPE_UP;
	settings.iTriggerLevel = 0;
	settings.iBufferSize = 4096;

	upcoming_settings = settings;
	
	frame_position = 0;
	triggered = false;
	bManualTrigger = false;

	stream.buffer = NULL;
}

DirectSoundScope::~DirectSoundScope()
{
}

void DirectSoundScope::init_waveformatex (PCMWAVEFORMAT *pcmwf, 
			       unsigned clock_rate,
			       unsigned channel_count)
{
    memset(pcmwf, 0, sizeof(PCMWAVEFORMAT)); 
    pcmwf->wf.wFormatTag = WAVE_FORMAT_PCM; 
	pcmwf->wf.nChannels = channel_count;
	pcmwf->wf.nSamplesPerSec = clock_rate;
	pcmwf->wf.nBlockAlign = channel_count * BYTES_PER_SAMPLE;
	pcmwf->wf.nAvgBytesPerSec = clock_rate * channel_count * BYTES_PER_SAMPLE;
	pcmwf->wBitsPerSample = BITS_PER_SAMPLE;
}

int DirectSoundScope::init_stream(struct ThreadComm::DsoundStream *strm, 
	   unsigned clock_rate, unsigned channel_count, unsigned samples_per_frame)
{
    HRESULT hr;
    PCMWAVEFORMAT pcmwf; 
    DSCBUFFERDESC dscbdesc;
	DSBPOSITIONNOTIFY dsPosNotify[BUFFER_COUNT];
    unsigned bytes_per_frame;

	int dev_id = -1;
	if (asDeviceName == "")
		dev_id = 0;

	dev_count = 0;
	/* Enumerate sound capture devices */
	hr = DirectSoundCaptureEnumerate(DSEnumCallback, NULL);
	if (FAILED(hr))
	{
		log("DirectSoundCaptureEnumerate failed with status %d", hr);
		return -1;
	}

	if (dev_count == 0)
	{
		log("No capture device found");
		return -1;
	}

	if (dev_id < 0)
	{
		for (unsigned int i=0; i<dev_count; i++) {
			if (asDeviceName == dev_info[i].name.c_str()) {
				dev_id = i;
				break;
			}
		}
	}

	if (dev_id < 0 || dev_id >= (int)dev_count)
	{
		log("Could not find selected device (%s)", asDeviceName.c_str());
		return -1;
	}

	asDeviceName = dev_info[dev_id].name.c_str();

	hr = DirectSoundCaptureCreate(dev_info[dev_id].lpGuid, &strm->lpDs, NULL);
	if (FAILED(hr))
	{
		log("DirectSoundCaptureCreate failed with status %d", hr);
		return 1;
	}

	// Init wave format to initialize buffer
	init_waveformatex( &pcmwf, clock_rate, channel_count);
	bytes_per_frame = samples_per_frame * BYTES_PER_SAMPLE * channel_count;
	stream.bytes_per_frame = bytes_per_frame;

	// Setup capture buffer
	memset(&dscbdesc, 0, sizeof(DSCBUFFERDESC));
	dscbdesc.dwSize = sizeof(DSCBUFFERDESC);
	dscbdesc.dwFlags = DSCBCAPS_WAVEMAPPED ;
	dscbdesc.dwBufferBytes = BUFFER_COUNT * bytes_per_frame;
	dscbdesc.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;

	hr = IDirectSoundCapture_CreateCaptureBuffer(strm->lpDs, &dscbdesc, &strm->lpDsBuffer, NULL);
	if (FAILED(hr))
	{
		log("IDirectSoundCapture_CreateCaptureBuffer failed with status %d", hr);
		return 2;
	}

	// Create event for notification.
	strm->hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	if (strm->hEvent == NULL)
	{
		log("Failed to create notification event");
		return 3;
	}

	// Setup notifications for recording.
	hr = IDirectSoundCaptureBuffer_QueryInterface(strm->lpDsBuffer,	IID_IDirectSoundNotify, (LPVOID *)&strm->lpDsNotify);
	if (FAILED(hr))
	{
		log("IDirectSoundCaptureBuffer_QueryInterface failed with status %d", hr);
		return 4;
	}
    
	for (unsigned int i=0; i<BUFFER_COUNT; i++)
	{
		dsPosNotify[i].dwOffset = i * bytes_per_frame;
		dsPosNotify[i].hEventNotify = strm->hEvent;
    }
    
	hr = IDirectSoundNotify_SetNotificationPositions( strm->lpDsNotify,
						      BUFFER_COUNT, 
						      dsPosNotify);
    if (FAILED(hr))
	{
		log("IDirectSoundNotify_SetNotificationPositions failed with status %d", hr);
		return 5;
	}

	hr = IDirectSoundCaptureBuffer_GetCurrentPosition(strm->lpDsBuffer, NULL, &strm->dwBytePos);
    if (FAILED(hr))
	{
		log("IDirectSoundCaptureBuffer_GetCurrentPosition failed with status %d", hr);
		return 6;
	}

	strm->dwDsBufferSize = BUFFER_COUNT * bytes_per_frame;

	if (strm->buffer)
	{
		free(strm->buffer);
		strm->buffer = NULL;
	}
	strm->buffer = malloc(bytes_per_frame);
	if (strm->buffer == NULL)
	{
		log("Failed to alloc stream buffer (no memory?)", hr);
		return 7;
	}

	log("DirectSound capture \"%s\" initialized (clock_rate=%d, "
		  "channel_count=%d, samples_per_frame=%d (%dms))",
		  dev_info[dev_id].name.c_str(),
		  clock_rate, channel_count, samples_per_frame,
	      samples_per_frame * 1000 / clock_rate);

	hr = IDirectSoundCaptureBuffer_Start(strm->lpDsBuffer,
					     DSCBSTART_LOOPING );
	if (FAILED(hr))
	{
		log("IDirectSoundCaptureBuffer_Start failed with status %d", hr);
		return 6;
	}

    return 0;
}


void DirectSoundScope::log(char *lpData, ...)
{
    static char buf[256]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int DirectSoundScope::Connect(void)
{
	assert(callbackTrigger && callbackData && callbackLog && callbackConnect);

	if (conn_state != DEVICE_DISCONNECTED)
		Disconnect();

	int status = init_stream( &stream, initialPluginSettings.samplingSpeed, CHANNEL_CNT, initialPluginSettings.samplingSpeed/10);

    if (status != 0)
    {
		SetConnectionState(DEVICE_DISCONNECTED, "Could not init capture device.");
		return 1;
    }
    else
    {
		AnsiString msg;
		msg.sprintf("Connected: %s", asDeviceName.c_str());
		SetConnectionState(DEVICE_CONNECTED, msg.c_str());
		threadComm = new ThreadComm();
		threadComm->stream = &stream;
		threadComm->callbackLog = callbackLog;
		threadComm->callbackDataRcv = DataRcvCb;
		threadComm->callbackConnect = callbackConnect;
		threadComm->Execute();
	}
    return 0;
}

int DirectSoundScope::Disconnect(void)
{
	SetConnectionState(DEVICE_DISCONNECTING, "Disconnecting...");
	if (threadComm)
	{
		threadComm->connected = false;
		Sleep(1000);
		delete threadComm;
		threadComm = NULL;
	}

	if (stream.lpDsNotify) {
		IDirectSoundNotify_Release(stream.lpDsNotify);
		stream.lpDsNotify = NULL;
	}

	if (stream.hEvent) {
		CloseHandle(stream.hEvent);
		stream.hEvent = NULL;
	}

	if (stream.lpDsBuffer) {
		IDirectSoundCaptureBuffer_Release(stream.lpDsBuffer);
		stream.lpDsBuffer = NULL;
	}

	if (stream.lpDs) {
		IDirectSoundCapture_Release(stream.lpDs);
		stream.lpDs = NULL;
	}

	if (stream.buffer)
	{
		free(stream.buffer);
		stream.buffer = NULL;
	}

	SetConnectionState(DEVICE_DISCONNECTED, "Disconnected");
	return 0;
}

int DirectSoundScope::ManualTrigger(void)
{
	if (!threadComm)
		return -1;
	if (conn_state != DEVICE_CONNECTED)
		return -2;
	bManualTrigger = true;
	return 0;
}

int DirectSoundScope::Run(bool state)
{
	bRunning = state;
	if (!bRunning)
	{
        bManualTrigger = false;
		lastTriggerDistance = 0;
		frame_position = 0;
	}
	return 0;
}

int DirectSoundScope::SetTriggerMode(enum E_TRIGGER_MODE m)
{
	mode = m;
	if (mode == TRIGGER_MODE_AUTO)
		lastTriggerDistance = 0;	// reset
	return 0;
}

int DirectSoundScope::SetTriggerType(int channel, enum E_TRIGGER_TYPE type)
{
	if (!threadComm)
		return -1;
	if (type == TRIGGER_SLOPE_UP)
	{
		upcoming_settings.eTriggerType = type;
	}
	else if (type == TRIGGER_SLOPE_DOWN)
	{
		upcoming_settings.eTriggerType = type;
	}
	else
		return -2;
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -3;
	upcoming_settings.iTriggerSrcChannel = channel;
	return 0;
}

int DirectSoundScope::SetCouplingType(int channel, enum E_COUPLING_TYPE type)
{
	return -1;
}

int DirectSoundScope::SetBufferSize(int size)
{
	if (!threadComm)
		return -1;
	upcoming_settings.iBufferSize = size;
	return 0;
}

int DirectSoundScope::SetSensitivity(int channel, int id)
{
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -1;
	/** \note Assuming that sensitivity is updated manually
	*/
	return 0;
}

int DirectSoundScope::SetSampling(int id)
{
	if (id < 0 || id > 10)
		return -1;
	upcoming_settings.iSamplingPeriod = (1 << id);
	return 0;
}

int DirectSoundScope::SetTriggerLevel(int level)
{
	if (level < -100 || level > 100)
		return -1;
	upcoming_settings.iTriggerLevel = level;
	return 0;
}

static unsigned char* memsearch(const unsigned char* hay, int haylen, const unsigned char* needle, int needlelen)
{
	for (int i=0; i<=haylen-needlelen; i++)
	{
		if (memcmp(hay+i, needle, needlelen) == 0)
			return (unsigned char*)(hay+i);
	}
	return NULL;
}

void DirectSoundScope::DataRcvCb(short data[], unsigned int samples)
{
	log("DataRcv %d samples", samples);
	//ConnectCb(DEVICE_CONNECTED, txtbuf);
	if (conn_state == DEVICE_CONNECTED)
	{
		if (samples == 0)
			return;
		if (samples % CHANNEL_CNT)
		{
			log("Unexpected: uneven data count!");
			return;
		}

		if (frame_position == 0)
		{
			settings = upcoming_settings;
			decimator_ch1.SetFactor(settings.iSamplingPeriod);
			decimator_ch2.SetFactor(settings.iSamplingPeriod);
		}

		// reduce bit number to 15 to add space for calibration
		for (unsigned int i=0; i<samples; i++)
		{
			data[i] /= 2;
		}

		// demux channels
		unsigned int pos = 0;
		for (unsigned int i=0; i<samples; i+=2)
		{
			samples_ch1[pos++] = data[i];
		}
		pos = 0;
		for (unsigned int i=1; i<samples; i+=2)
		{
			samples_ch2[pos++] = data[i];
		}

		// calculate average sample values for calibration purposes
		long long sumA = 0;
		for (unsigned int i=0; i<samples/2; i++)
		{
			sumA += samples_ch1[i];
		}
		averageA = sumA / (samples/2);
		long long sumB = 0;
		for (unsigned int i=0; i<samples/2; i++)
		{
			sumB += samples_ch2[i];
		}
		averageB = sumB / (samples/2);

		// add offset
		for (unsigned int i=0; i<samples/2; i++)
		{
			samples_ch1[i] += pluginSettings.offset;
		}
		for (unsigned int i=0; i<samples/2; i++)
		{
			samples_ch2[i] += pluginSettings.offset;
		}

		// decimate samples if needed
		int iSamplesCnt = decimator_ch1.Process(samples_ch1, samples/2, output_buffer_ch1);
		decimator_ch2.Process(samples_ch2, samples/2, output_buffer_ch2);
		if (iSamplesCnt <= 0)
			return;

		short* samples_ch1_p = output_buffer_ch1;
		short* samples_ch2_p = output_buffer_ch2;

		if (bRunning == false && bManualTrigger == false)
		{
			frame_position = 0;
			return;
		}

		if (frame_position == 0)
		{
			int trigger_pos;
			if (settings.iTriggerSrcChannel == 0)
				trigger_pos = TestTrigger(output_buffer_ch1, iSamplesCnt);
			else
				trigger_pos = TestTrigger(output_buffer_ch2, iSamplesCnt);

			if (trigger_pos < 0)
			{
				// ignore this data portion for now
				return;
			}
			else
			{
				samples_ch1_p += trigger_pos;
				samples_ch2_p += trigger_pos;
				iSamplesCnt -= trigger_pos;
				log("callbackTrigger, no earlier data");
				callbackTrigger();
			}
		}
		log("callbackData samples after decimation, per channel: %d, %d B", iSamplesCnt, iSamplesCnt * sizeof(output_buffer_ch1[0]) * 2);
		if (iSamplesCnt + frame_position < settings.iBufferSize)
		{
            int offset = 0;
			if (settings.pChannelEnable[0])
			{
				memcpy(output_buffer_concat, samples_ch1_p, iSamplesCnt * sizeof(output_buffer_ch1[0]));
				offset = iSamplesCnt;
			}
			if (settings.pChannelEnable[1])
			{
				memcpy((char*)output_buffer_concat + offset * sizeof(output_buffer_ch1[0]),
					samples_ch2_p, iSamplesCnt * sizeof(output_buffer_ch2[0]));
				offset += iSamplesCnt;
			}
			//log("L%03d | Ch1: %d, %d | Ch2: %d, %d", __LINE__, samples_ch1_p[0], samples_ch1_p[1], samples_ch2_p[0], samples_ch2_p[1]);
			callbackData(output_buffer_concat, offset * sizeof(output_buffer_ch1[0]));
			frame_position += iSamplesCnt;
		}
		else
		{
			if (settings.iBufferSize-frame_position > 0)
			{
				assert (settings.iBufferSize > frame_position);
				int iSamplesConsumed = settings.iBufferSize-frame_position;

				int offset = 0;
				if (settings.pChannelEnable[0])
				{
					memcpy(output_buffer_concat, samples_ch1_p, iSamplesConsumed * sizeof(output_buffer_ch1[0]));
					offset = iSamplesConsumed;
				}
				if (settings.pChannelEnable[1])
				{
					memcpy((char*)output_buffer_concat + offset * sizeof(output_buffer_ch1[0]),
						samples_ch2_p, iSamplesConsumed * sizeof(output_buffer_ch2[0]));
					offset += iSamplesConsumed;
				}
				callbackData(output_buffer_concat, offset * sizeof(output_buffer_ch1[0]));


				frame_position += iSamplesConsumed;

				iSamplesCnt -= iSamplesConsumed;
				memmove(output_buffer_ch1, output_buffer_ch1+iSamplesConsumed,
					iSamplesCnt * sizeof(output_buffer_ch1[0]));
				memmove(output_buffer_ch2, output_buffer_ch2+iSamplesConsumed,
					iSamplesCnt * sizeof(output_buffer_ch2[0]));
				frame_position = iSamplesCnt;
			}
			bManualTrigger = false;			
			if (mode != TRIGGER_MODE_CONTINUOUS)
			{
				frame_position = 0;
				settings = upcoming_settings;
				decimator_ch1.SetFactor(settings.iSamplingPeriod);
				decimator_ch2.SetFactor(settings.iSamplingPeriod);
			}
			else
			{
				settings = upcoming_settings;
				decimator_ch1.SetFactor(settings.iSamplingPeriod);
				decimator_ch2.SetFactor(settings.iSamplingPeriod);
				if (iSamplesCnt > 0)
				{
					log("callbackTrigger, %d samples left", iSamplesCnt);
					callbackTrigger();

					int offset = 0;
					if (settings.pChannelEnable[0])
					{
						memcpy(output_buffer_concat, output_buffer_ch1, iSamplesCnt * sizeof(output_buffer_ch1[0]));
						offset = iSamplesCnt;
					}
					if (settings.pChannelEnable[1])
					{
						memcpy((char*)output_buffer_concat + offset * sizeof(output_buffer_ch1[0]),
							output_buffer_ch2, iSamplesCnt * sizeof(output_buffer_ch2[0]));
						offset += iSamplesCnt;
					}
					//log("L%03d | Ch1: %d, %d | Ch2: %d, %d", __LINE__, samples_ch1_p[0], samples_ch1_p[1], samples_ch2_p[0], samples_ch2_p[1]);
					callbackData(output_buffer_concat, offset * sizeof(output_buffer_ch1[0]));
				}
				if (frame_position == 0)
					bManualTrigger = true;
			}

			if (mode == TRIGGER_MODE_SINGLE)
				bRunning = false;
		}
	}
}

void DirectSoundScope::SetConnectionState(enum E_CONNECTION_STATE state, const char* const descr)
{
	conn_state = state;
	if (callbackConnect)
		callbackConnect(state, (char*)descr);
	log("Connection state changed to %d. %s", (int)state, descr);
}

int DirectSoundScope::TestTrigger(short* output_buffer, int len)
{
	int pos = -1;
	int level = 327.68f * (float)settings.iTriggerLevel;	//-100...100 => -x...x range
	if (bManualTrigger || mode == TRIGGER_MODE_CONTINUOUS)
	{
		pos = 0;
	}
	else
	{
		switch (settings.eTriggerType)
		{
		case TRIGGER_SLOPE_UP:
			for (int i=0; i<len-6; i++)
			{
				if (output_buffer[i] < level && output_buffer[i+1] <= level && output_buffer[i+2] <= level &&
					output_buffer[i+3] >= level && output_buffer[i+4] >= level && output_buffer[i+5] > level)
				{
					//log("i = %d, len = %d, %d %d %d %d %d %d", i, len, output_buffer[i], output_buffer[i+1], output_buffer[i+2],
					//	output_buffer[i+3], output_buffer[i+4], output_buffer[i+5]);
					pos = i;
					break;
				}
			}
			break;
		case TRIGGER_SLOPE_DOWN:
			for (int i=0; i<len-6; i++)
			{
				if (output_buffer[i] > level && output_buffer[i+1] >= level && output_buffer[i+2] >= level &&
					output_buffer[i+3] <= level && output_buffer[i+4] <= level && output_buffer[i+5] < level)
				{
					//log("i = %d, len = %d, %d %d %d %d %d %d", i, len, output_buffer[i], output_buffer[i+1], output_buffer[i+2],
					//	output_buffer[i+3], output_buffer[i+4], output_buffer[i+5]);
					pos = i;
					break;
				}
			}
			break;
		default:
			break;
		}
	}
	if (mode == TRIGGER_MODE_AUTO)
	{
		if (pos == -1)
		{
			lastTriggerDistance += len;
			// generate auto-trigger when there is no trigger for the equivalent time of 3 frames
			if (lastTriggerDistance >= 3 * upcoming_settings.iBufferSize)
			{
				pos = 0;
				lastTriggerDistance = 0;
			}
		}
		else
		{
			lastTriggerDistance = 0;
		}
	}
	return pos;
}

void DirectSoundScope::GetCaptureDevicesList(std::vector<std::string> &devices)
{
	devices.clear();
	for (unsigned int index=0; index<dev_count; ++index)
	{
		devices.push_back(dev_info[index].name);
	}
}

void DirectSoundScope::SetDevice(AnsiString name)
{
	asDeviceName = name;
}

int DirectSoundScope::GetAverageSample(int channel)
{
	if (channel == 0)
		return averageA;
	else
		return averageB;
}

int DirectSoundScope::EnableChannel(int channel, bool state)
{
	upcoming_settings.pChannelEnable[channel] = state;
	return 0;
}

